--
-- DE2 (Cyclone-II) Entity for Interactive Project Game
-- Authors:
--      Abdulhamid Ghandour
--      Thomas John
--      Jaime Peretzman
--      Bharadwaj Vellore
--
-- Desc:
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity visionsystem is

port(
clk : in std_logic;

pixel_data : in unsigned (9 downto 0);
valid_green : in std_logic;

endofrow : in std_logic;
endofframe : in std_logic;

threshold : in unsigned (9 downto 0);

xout : out unsigned (15 downto 0);
yout : out unsigned (15 downto 0);

no_detect : out std_logic;

led0, led1, led2, led3, led4, led5, led6, led7 : out std_logic
);

end visionsystem;


architecture rtl of visionsystem is

constant FINGER_WIDTH : unsigned (15 downto 0) := x"002A";
constant WIDTH_SAMPLE_INTERVAL : unsigned (15 downto 0) := x"0014";
constant BOUNDARY_TOLERANCE : unsigned (15 downto 0) := x"0003";
constant SAME_EDGE_TOLERANCE : unsigned (15 downto 0) := x"000A";

signal led0_buff, led1_buff, led2_buff, led3_buff, led4_buff, led5_buff, led6_buff, led7_buff : std_logic;

signal xcount : unsigned (15 downto 0);
signal ycount : unsigned (15 downto 0);

--------------------EXTREMETIES--------------------------------
signal topx    : unsigned (15 downto 0) := (others => '0');
signal topy    : unsigned (15 downto 0) := (others => '1');
signal bottomx : unsigned (15 downto 0) := (others => '0');
signal bottomy : unsigned (15 downto 0) := (others => '0');
signal leftx   : unsigned (15 downto 0) := (others => '1');
signal lefty   : unsigned (15 downto 0) := (others => '0');
signal rightx  : unsigned (15 downto 0) := (others => '0');
signal righty  : unsigned (15 downto 0) := (others => '0');

signal filtered_rightx  : unsigned (15 downto 0) := (others => '0');
signal filtered_righty  : unsigned (15 downto 0) := (others => '0');
signal row_max          : unsigned (15 downto 0) := (others => '0');
signal prev_row_max     : unsigned (15 downto 0) := (others => '0');
signal prev_row_max2     : unsigned (15 downto 0) := (others => '0');
signal prev_row_max3     : unsigned (15 downto 0) := (others => '0');
-----------------------WIDTH------------------------------------
signal topwidth_start : unsigned (15 downto 0) := (others => '1');
signal topwidth_end   : unsigned (15 downto 0) := (others => '0');

signal prev_bottomy : unsigned (15 downto 0) := (others => '0');
signal bottomwidth_start : unsigned (15 downto 0) := (others => '0');
signal bottomwidth_end   : unsigned (15 downto 0) := (others => '0');

type unsignedarray_type is array(0 to 19) of unsigned (15 downto 0); 
signal bottomwidth_start_array : unsignedarray_type := (others => x"FFFF");
signal bottomwidth_end_array   : unsignedarray_type := (others => x"0000");
signal bottomwidth_recorded    : std_logic := '0';
------------------------SCREEN---------------------------------------
signal maxx : unsigned (15 downto 0) := (others => '0');
signal maxy : unsigned (15 downto 0) := (others => '0');
--------------------------------------------------------------------

begin

xcounter : process(clk)
begin
if rising_edge(clk) then
if endofrow = '1' or endofframe = '1' then
xcount <= (others => '0');
maxx <= xcount - 1;
elsif valid_green = '1' then
xcount <= xcount + 1;
end if;
end if;
end process xcounter;

ycounter : process(clk)
begin
if rising_edge(clk) then
if endofframe = '1' then
ycount <= (others => '0');
maxy <= ycount - 1;
elsif endofrow = '1' then
ycount <= ycount + 1;
end if;
end if;
end process ycounter;

-- Extract relevant data as image information comes in:
-- -- TOP EXTREME
-- -- BOTTOM EXTREME
-- -- LEFT EXTREME
-- -- RIGHT EXTREME
-- -- HORIZONTAL WIDTH AT A CONSTANT DISTANCE BELOW TOP EXTREME
-- -- HORIZONTAL WIDTH AT A CONSTANT DISTANCE ABOVE BOTTOM EXTREME

data_extraction : process (clk)
variable filterx_temp : unsigned (15 downto 0);
variable filter_sum_temp : unsigned (15 downto 0);
begin
if rising_edge(clk) then

if valid_green = '1' then
if pixel_data < threshold then

-- TOP
if ycount <= topy then
topy <= ycount;
topx <= xcount;
end if;

-- BOTTOM
if ycount >= bottomy then
bottomy <= ycount;
bottomx <= xcount;
end if;

-- LEFT
if xcount <= leftx then
leftx <= xcount;
lefty <= ycount;
end if;
    
-- RIGHT
if xcount >= rightx then
rightx <= xcount;
righty <= ycount;
end if;

-- FILTERED RIGHT
if xcount > row_max then
row_max <= xcount;
end if;

if not (topy = x"FFFF") and ycount = topy + WIDTH_SAMPLE_INTERVAL then
if topwidth_start = x"FFFF" then
topwidth_start <= xcount;
end if;
        end if;

        if bottomwidth_start_array(0) = x"FFFF" then
bottomwidth_start_array(0) <= xcount;
end if;

elsif pixel_data >= threshold then
if not (topy = x"FFFF") and ycount = topy + WIDTH_SAMPLE_INTERVAL then
if not (topwidth_start = x"FFFF") and topwidth_end = 0 then
topwidth_end <= xcount;
end if;
        end if;

        if not (bottomwidth_start_array(0) = x"FFFF") and bottomwidth_end_array(0) = 0 then
bottomwidth_end_array(0) <= xcount;
end if;

end if;  -- pixel_data < threshold
end if;  -- valid_green = '1'

if endofrow = '1' then

	prev_row_max3 <= prev_row_max2;
	prev_row_max2 <= prev_row_max;
	prev_row_max <= row_max;
	row_max <= (others => '0');
	filter_sum_temp := (row_max + prev_row_max + prev_row_max2 + prev_row_max3);
	filterx_temp := ("00" & filter_sum_temp(15 downto 2));
	
	if filterx_temp > filtered_rightx then
		filtered_rightx <= filterx_temp;
		filtered_righty <= ycount;
	end if;
	
	bottomwidth_start_array(0) <= x"FFFF";
	bottomwidth_start_array(1) <= bottomwidth_start_array(0);
	bottomwidth_start_array(2) <= bottomwidth_start_array(1);
	bottomwidth_start_array(3) <= bottomwidth_start_array(2);
	bottomwidth_start_array(4) <= bottomwidth_start_array(3);
	bottomwidth_start_array(5) <= bottomwidth_start_array(4);
	bottomwidth_start_array(6) <= bottomwidth_start_array(5);
	bottomwidth_start_array(7) <= bottomwidth_start_array(6);
	bottomwidth_start_array(8) <= bottomwidth_start_array(7);
	bottomwidth_start_array(9) <= bottomwidth_start_array(8);
	bottomwidth_start_array(10) <= bottomwidth_start_array(9);
	bottomwidth_start_array(11) <= bottomwidth_start_array(10);
	bottomwidth_start_array(12) <= bottomwidth_start_array(11);
	bottomwidth_start_array(13) <= bottomwidth_start_array(12);
	bottomwidth_start_array(14) <= bottomwidth_start_array(13);
	bottomwidth_start_array(15) <= bottomwidth_start_array(14);
	bottomwidth_start_array(16) <= bottomwidth_start_array(15);
	bottomwidth_start_array(17) <= bottomwidth_start_array(16);
	bottomwidth_start_array(18) <= bottomwidth_start_array(17);
	bottomwidth_start_array(19) <= bottomwidth_start_array(18);
	
	bottomwidth_end_array(0) <= x"0000";
	bottomwidth_end_array(1) <= bottomwidth_end_array(0);
	bottomwidth_end_array(2) <= bottomwidth_end_array(1);
	bottomwidth_end_array(3) <= bottomwidth_end_array(2);
	bottomwidth_end_array(4) <= bottomwidth_end_array(3);
	bottomwidth_end_array(5) <= bottomwidth_end_array(4);
	bottomwidth_end_array(6) <= bottomwidth_end_array(5);
	bottomwidth_end_array(7) <= bottomwidth_end_array(6);
	bottomwidth_end_array(8) <= bottomwidth_end_array(7);
	bottomwidth_end_array(9) <= bottomwidth_end_array(8);
	bottomwidth_end_array(10) <= bottomwidth_end_array(9);
	bottomwidth_end_array(11) <= bottomwidth_end_array(10);
	bottomwidth_end_array(12) <= bottomwidth_end_array(11);
	bottomwidth_end_array(13) <= bottomwidth_end_array(12);
	bottomwidth_end_array(14) <= bottomwidth_end_array(13);
	bottomwidth_end_array(15) <= bottomwidth_end_array(14);
	bottomwidth_end_array(16) <= bottomwidth_end_array(15);
	bottomwidth_end_array(17) <= bottomwidth_end_array(16);
	bottomwidth_end_array(18) <= bottomwidth_end_array(17);
	bottomwidth_end_array(19) <= bottomwidth_end_array(18);
	
	if not (bottomy = 0) and bottomy = prev_bottomy and bottomwidth_recorded = '0' then
		bottomwidth_start <= bottomwidth_start_array(19);
		bottomwidth_end   <= bottomwidth_end_array(19);
		bottomwidth_recorded <= '1';
	end if;
	
	prev_bottomy <= bottomy;

end if;

-- RESET AT END OF FRAME
if endofframe = '1' then
topx    <=  (others => '0');
topy    <=  (others => '1');
bottomx <=  (others => '0');
bottomy <=  (others => '0');
leftx   <= (others => '1');
lefty   <= (others => '0');
rightx  <= (others => '0');
righty  <= (others => '0');

prev_row_max <= (others => '0');
row_max <= (others => '0');
filtered_rightx <= (others => '0');
filtered_righty <= (others => '0');

topwidth_start <= (others => '1');
topwidth_end   <= (others => '0');

prev_bottomy <= (others => '0');
bottomwidth_start <= (others => '0');
bottomwidth_end   <= (others => '0'); 
bottomwidth_start_array <= (others => x"FFFF");
bottomwidth_end_array   <= (others => x"0000");
bottomwidth_recorded <= '0';
end if;

end if;

end process data_extraction;

output : process(clk)
variable top_on_edge, bottom_on_edge, left_on_edge, right_on_edge,
         topentry, bottomentry, leftentry, rightentry : integer := 0;
variable xdiff, ydiff : unsigned (15 downto 0); 
begin
if rising_edge(clk) then
-----------------------------------------------------
if endofframe = '1' then

led0_buff <= '0';
led1_buff <= '0';
led2_buff <= '0';
led3_buff <= '0';
led4_buff <= '0';
led5_buff <= '0';
led6_buff <= '0';
led7_buff <= '0';

if topy = x"FFFF" and rightx = 0 and leftx = x"FFFF" and bottomy = 0 then
no_detect <= '1';
else
no_detect <= '0';
end if;

if topy <= BOUNDARY_TOLERANCE or topx <= BOUNDARY_TOLERANCE or topx > maxx - BOUNDARY_TOLERANCE then
top_on_edge := 1;
else
top_on_edge := 0;
end if;
if topy = 0 then
topentry := 1;
else
topentry := 0;
end if;

if bottomy > maxy - BOUNDARY_TOLERANCE or bottomx <= BOUNDARY_TOLERANCE or bottomx > maxx - BOUNDARY_TOLERANCE then
bottom_on_edge := 1;
else
bottom_on_edge := 0;
end if;
if bottomy = maxy then
bottomentry := 1;
else
bottomentry := 0;
end if;

if leftx <= BOUNDARY_TOLERANCE or lefty <= BOUNDARY_TOLERANCE or lefty > maxy - BOUNDARY_TOLERANCE then
left_on_edge := 1;
else
left_on_edge := 0;
end if;
if leftx = 0 then
leftentry := 1;
else
leftentry := 0;
end if;

if rightx > maxx - BOUNDARY_TOLERANCE or righty <= BOUNDARY_TOLERANCE or righty > maxy - BOUNDARY_TOLERANCE then
right_on_edge := 1;
else
right_on_edge := 0;
end if;
if rightx = maxx then
rightentry := 1;
else
rightentry := 0;
end if;

--------------------------------------------------
if (top_on_edge + bottom_on_edge + right_on_edge + left_on_edge) = 1 then
if topentry = 1 then
xout <= bottomx;
yout <= bottomy;
led5_buff <= '1';
end if;

if bottomentry = 1 then
xout <= topx;
yout <= topy;
led6_buff <= '1';
end if;

if leftentry = 1 then
xout <= filtered_rightx;
yout <= filtered_righty;
led0_buff <= '1';
end if;

if rightentry = 1 then
xout <= leftx;
yout <= lefty;
end if;
-- - - - - -  - - - - - - - - - - - - - - - - - - - - - - - -
elsif (top_on_edge + bottom_on_edge + right_on_edge + left_on_edge) = 3 then
if top_on_edge = 0 then
xout <= topx;
yout <= topy;
end if;

if bottom_on_edge = 0 then
xout <= bottomx;
yout <= bottomy;
end if;

if left_on_edge = 0 then
xout <= leftx;
yout <= lefty;
end if;

if right_on_edge = 0 then
xout <= filtered_rightx;
yout <= filtered_righty;
end if;
-- - - - - - - - - - - - - - - - - - - - - - - - - - -	
elsif top_on_edge + bottom_on_edge + right_on_edge + left_on_edge = 2 then

--****************TOP LEFT*******************
if top_on_edge = 1 and left_on_edge = 1 then

if(bottomx > rightx) then
xdiff := bottomx - rightx;
else
xdiff := rightx - bottomx;
end if;

if(bottomy > righty) then
ydiff := bottomy - righty;
else
ydiff := righty - bottomy;
end if;

if (xdiff < SAME_EDGE_TOLERANCE and ydiff < SAME_EDGE_TOLERANCE) then
xout <= filtered_rightx;
yout <= filtered_righty;
else
if bottomwidth_end - bottomwidth_start >= FINGER_WIDTH then
xout <= filtered_rightx;
yout <= filtered_righty;
led1_buff <= '1';
else
xout <= filtered_rightx;
yout <= filtered_righty;
led2_buff <= '1';
end if;
end if;

--****************TOP RIGHT*******************
elsif top_on_edge = 1 and right_on_edge = 1 then

if(bottomx > leftx) then
xdiff := bottomx - leftx;
else
xdiff := leftx - bottomx;
end if;

if(bottomy > lefty) then
ydiff := bottomy - lefty;
else
ydiff := lefty - bottomy;
end if;

if (xdiff < SAME_EDGE_TOLERANCE and ydiff < SAME_EDGE_TOLERANCE) then
xout <= leftx;
yout <= lefty;
else
if bottomwidth_end - bottomwidth_start >= FINGER_WIDTH then
xout <= leftx;
yout <= lefty;
else
xout <= leftx;
yout <= lefty;
end if;
end if;

--****************BOTTOM LEFT*******************
elsif bottom_on_edge = 1 and left_on_edge = 1 then

if(topx > rightx) then
xdiff := topx - rightx;
else
xdiff := rightx - topx;
end if;

if(topy > righty) then
ydiff := topy - righty;
else
ydiff := righty - topy;
end if;

if (xdiff < SAME_EDGE_TOLERANCE and ydiff < SAME_EDGE_TOLERANCE) then
xout <= rightx;
yout <= righty;
else
if topwidth_end - topwidth_start < FINGER_WIDTH then
xout <= topx;
yout <= topy;
led3_buff <= '1';
else
xout <= rightx;
yout <= righty;
led4_buff <= '1';
end if;
end if;

--****************BOTTOM RIGHT*******************
elsif bottom_on_edge = 1 and right_on_edge = 1 then

if(topx > leftx) then
xdiff := topx - leftx;
else
xdiff := leftx - topx;
end if;

if(topy > lefty) then
ydiff := topy - lefty;
else
ydiff := lefty - topy;
end if;

if (xdiff < SAME_EDGE_TOLERANCE and ydiff < SAME_EDGE_TOLERANCE) then
xout <= leftx;
yout <= lefty;
else
if topwidth_end - topwidth_start < FINGER_WIDTH then
xout <= topx;
yout <= topy;
else
xout <= leftx;
yout <= lefty;
end if;
end if;
end if;
--************************************************	

end if;

--------------------------------------------------
end if; -- End of Frame
end if;
end process output;

process(clk)
begin
led0 <= led0_buff;
led1 <= led1_buff;
led2 <= led2_buff;
led3 <= led3_buff;
led4 <= led4_buff;
led5 <= led5_buff;
led6 <= led6_buff;
led7 <= led7_buff;
end process;

end rtl;
